/*===============================================================================*\
|######################## 		  [PPSSPP FX 1.00]   	 #########################|
|########################			BY ASMODEAN	     	 #########################|
||																				 ||
||		  This program is free software; you can redistribute it and/or			 ||
||		  modify it under the terms of the GNU General Public License			 ||
||		  as published by the Free Software Foundation; either version 2		 ||
||		  of the License, or (at your option) any later version.				 ||
||																				 ||
||		  This program is distributed in the hope that it will be useful,		 ||
||		  but WITHOUT ANY WARRANTY; without even the implied warranty of		 ||
||		  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the			 ||
||		  GNU General Public License for more details. (c)2013					 ||
||																				 ||
|#################################################################################|
\*===============================================================================*/
#version 150 

/*------------------------------------------------------------------------------
						[DEFINITIONS & ON/OFF OPTIONS]
------------------------------------------------------------------------------*/

//---------------------------#[CHOOSE EFFECTS]#--------------------------------\\

//-#[ANTIALIASING TECHNIQUES]   [1=ON|0=OFF]  #READ: For best results, only enable one type of AA at one time.
#define FXAA_PASS					 0		//#High Quality Fast Approximate Anti Aliasing. Adapted from Timothy Lottes FXAA 3.11.
#define NFAA_PASS				 	 0		//#Normal Filter Anti Aliasing. Similar to 1x SMAA. (Subtle AA, with good edge detection).

//-#[TEXTURE FILTERING] 	    [1=ON|0=OFF]  #READ: For best results, don't use this with AA, use the built-in AA filter option. Only enable one type of filtering at one time.
#define BICUBIC_FILTERING			 0		//#BiCubic Fullscreen Filtering. Use an AA method or this. Not both.

//-#[LIGHTING & COLOUR] 	    [1=ON|0=OFF]  #READ: These can all be turned on & off independently of each other. (HDR = Bloom, Tonemapping, & Gamma Correction used together)
#define BLOOM_PASS				 0		//#HQ Bloom high pass [For best results: use bloom, gamma & tone mapping together].
#define SCENE_TONEMAPPING			 1		//#Scene Color Correction & Tonemapping [For best results: use bloom, gamma & tone mapping together].
#define GAMMA_CORRECTION			 1		//#Per-Channel Post Inverse Gamma Correction [For best results: use bloom, gamma & tone mapping together].
#define S_CURVE_CONTRAST			 1		//#S-Curve Scene Contrast Enhancement. Naturally adjusts contrast using S-curves.
#define TEXTURE_SHARPENING			 0		//#HQ Luma-Based Texture Sharpen, looks similar to a negative mip LOD Bias, enhances texture fidelity.
#define PIXEL_VIBRANCE			 	 1		//#Pixel Vibrance. Intelligently adjusts pixel vibrance depending on original saturation.
#define TOON_SHADING				 0		//#Cartoon-Style look, I'm still working on improving this, but it works fine for now. (interferes with other effects.)

/*------------------------------------------------------------------------------
						 [SHADER FX CONFIG OPTIONS]
------------------------------------------------------------------------------*/


//-[FXAA OPTIONS]-----------||
#define FxaaSubpixMax 1.00				//[0.10 to 1.00] Amount of subpixel aliasing removal. Higher values = (softer, more sub-pixel aliasing removal). Lower values = (sharper, less sub-pixel aliasing removal)
#define FxaaQuality 4					//[1|2|3|4] Overall FXAA quality preset.(LrQ/faster) 1 = LowQ, 2 = MediumQ, 3 = HighQ, 4 = UltraQ (HrQ/slower).
#define DebugFxaa 0						//[0 or 1] Visualize the pixel detection.

//-[NFAA OPTIONS]-----------||
#define FilterStrength 20.0f;			//Anti aliasing strength. The overall strength of the effect.
#define FilterSpread 0.20f;				//Filter spread width. For best results, keep a good balance between the strength and spread.
#define BilinearSampling 0				//Uses Bilinear filtering in the NFAA texture sampling.(Extra optional filtering.)
#define UseColor 1						//Uses color detection, instead of luma edge detection (more coverage, less accuracy, but better for 2D).
#define DebugNFAA 0						//[0 or 1] Visualize the edge detection.

//-[BICUBIC OPTIONS]--------||
#define FilterType BSpline				//[CatMullRom, BSpline, Triangular, Bell] Type of interpolation to use. Try them out, and use what you prefer.
#define OffsetAmount 0.0				//[0.0 to 1.5] Bicubic filtering pixel offset level. 0 is off.

//-[BLOOM OPTIONS]----------||
#define BloomType BlendScreen			//[BlendScreen, BlendOverlay, BlendAddLight] The type of blend for the bloom (Default: BlendScreen)
#define BloomPower 0.500f     			//[0.000 to 1.000] Strength of the bloom. blooms naturally, per environment. You will need to adjust for each bloom type.
#define BlurSpread 4.000f    			//[0.000 to 8.000] Width of the blooms' 'glow' offset. 0.000 = off. Recommended spread is 1.5 or 2 with BlendPower 2, and 2 or 4 with BlendPower 1. Scales with BloomPower. SharpenClamp affects this.
#define BloomMixType 1					//[1|2|3] The interpolation mix type between the bloom's color, and blur. (Default is 1) BloomPower/BlurSpread may need re-adjusting depending on type.
#define BlendPower 1					//[1 or 2] Strength of the blend glow. This only applies to the BlendScreen Bloom type. 1 is light, 2 is heavy. At 2, BlurSpread may need adjusting to compensate.

//-[TONEMAP OPTIONS]--------||
#define TonemapType 1					//[1|2|3] Type of tone mapping. 1 is natural(default), 2 is cinematic(dark), 3 is vivid(light). You might want to increase/decrease ToneAmount, or Bloom to compensate for diff types.
#define ToneAmount 0.80f				//[0.10 to 1.00] Gamma levels (Tonemapping/Shadow correction) lower: darker tones, higher: lighter tones.
#define Luma 100.00f					//Luma Correction(luminance correction) Higher values: lower luma, lower values: higher luma. Adjust by small amounts, eg: increments of 1
#define Exposure 100.00f				//White Correction (brightness correction) Higher values = more exposure, lower = less exposure. Adjust by small amounts, eg: increments of 1
#define WhitePoint 1.00f				//Whitepoint Avg (fog correction) Adjust by small amounts, eg: increments of 0.01

//-[CONTRAST OPTIONS]-------||
#define CurveType 0						//[0|1|2] Choose what to apply contrast to. 0 = Luma, 1 = Chroma, 2 = both Luma and Chroma. Default is 0 (Luma)
#define CurvesContrast 0.65f			//[-1.00 to 2.00] The amount of contrast you want. CurvesFormula 1 typically needs half the amount of 2, for the same strength.
#define CurvesFormula 2					//[1|2] The contrast s-curve you want to use. 1 is a harsher type. 2 is a softer type.

//-[SHARPEN OPTIONS]--------||
#define SharpenStrength 0.50f   		//[0.10 to 1.00] Strength of the texture sharpening effect.
#define SharpenClamp 0.012f				//[0.000 to 1.000] Clamps the maximum amount of sharpening each pixel recieves. Raise to reduce the clamping. High values can affect filtering.
#define OffsetBias 4.00f 				//[1.00 to 6.00] Sharpening offset bias. Use 1.0 for clean crisp sharpen, higher values for a deeper textured sharpen. High values can look odd(outline) at low IR.
#define DebugSharpen 0					//[0 or 1] Visualize the sharpening effect.

//-[VIBRANCE OPTIONS]-------||
#define Vibrance 0.50f 					//[-1.00 to 1.00] Intelligently saturates (or desaturates with negative values) pixels depending on their original saturation. 0.00 is original vibrance.

//-[TOON OPTIONS]-----------||
#define ToonVibrance 0.10f				//[-1.00 to 1.00] Vibrance adjustment. Higher values increase pixel vibrance, lower values reduce it. 
#define ToonStrength 2.00f				//[0.10 to 10.00] Strength of the toon outline effect.
#define ToonEdgeSlope 0.70f				//[0.10 to 2.00] Raise this to filter out fainter edges. You might need to increase the power to compensate.

//-[GAMMA OPTIONS]----------||
#define Gamma 2.2f  					//Lower values for more Gamma toning(darker), higher Values for brighter (2.2 correction is generally recommended)

//-[END OF USER OPTIONS]----||

/*------------------------------------------------------------------------------
							 [GLOBAL VARIABLES]
------------------------------------------------------------------------------*/

smooth in vec2 v_texcoord0;
uniform sampler2D sampler0;
uniform vec2 u_texelDelta;
smooth out vec4 gl_FragColor;

float RGBLuminance(vec3 color)
{
	vec3 lumCoeff = vec3(0.212656f, 0.715158f, 0.072186f);
	return dot(color.rgb, lumCoeff);
}

float TrueLuminance(vec3 color)
{
	float maxRGB;
	float minRGB;
	float r = color.x;
	float g = color.y;
	float b = color.z;

	if (r >= g) { maxRGB = r; }
	if (r >= b) { maxRGB = r; }
	if (g >= r) { maxRGB = g; }
	if (g >= b) { maxRGB = g; }
	if (b >= r) { maxRGB = b; }
	if (b >= g) { maxRGB = b; }

	if (r <= g) { minRGB = r; }
	if (r <= b) { minRGB = r; }
	if (g <= r) { minRGB = g; }
	if (g <= b) { minRGB = g; }
	if (b <= r) { minRGB = b; }
	if (b <= g) { minRGB = b; }

	float lumin = ((maxRGB + minRGB)/2);
	return lumin;
}

#define GammaCorrection(color, gamma) pow(color, gamma)
#define InverseGammaCorrection(color, gamma) pow(color, 1.0f/gamma)

/*------------------------------------------------------------------------------
                             [FXAA CODE SECTION]
------------------------------------------------------------------------------*/

#if (FXAA_PASS == 1)
#if (FxaaQuality == 4)
	#define FXAAEdgeThreshold (0.0312)
	#define FXAAEdgeThresholdMin (0.0156)
#endif
#if (FxaaQuality == 3)
	#define FXAAEdgeThreshold (0.125)
	#define FXAAEdgeThresholdMin (0.0312)
#endif
#if (FxaaQuality == 2)
	#define FXAAEdgeThreshold (0.166)
	#define FXAAEdgeThresholdMin (0.0625)
#endif
#if (FxaaQuality == 1)
	#define FXAAEdgeThreshold (0.250)
	#define FXAAEdgeThresholdMin (0.0833)
#endif

float FxaaLuma(vec4 rgba)
{
	return rgba.y;
}

vec4 PreFxaaPass(vec4 color, vec2 v_texcoord0)
{
	color = texture(sampler0, v_texcoord0);
	color.rgba = color.xyzw;
	color.xyz = pow(color.xyz, vec3(1.0 / 2.2));
	color.w = RGBLuminance(color.xyz);
	return color;
}

vec4 FxaaPixelShader(vec2 v_texcoord0, sampler2D tex, vec2 fxaaQualityRcpFrame, float fxaaQualitySubpix, float fxaaQualityEdgeThreshold, float fxaaQualityEdgeThresholdMin)
{
	vec2 posM;
	posM.x = v_texcoord0.x;
	posM.y = v_texcoord0.y;
	
	vec4 rgbyM = textureLod(tex, posM, 0.0);
	
	float lumaS = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2( 0, 1)));
	float lumaE = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2( 1, 0)));
	float lumaN = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2( 0,-1)));
	float lumaW = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2(-1, 0)));
	
	float maxSM = max(lumaS, rgbyM.y);
	float minSM = min(lumaS, rgbyM.y);
	float maxESM = max(lumaE, maxSM);
	float minESM = min(lumaE, minSM);
	float maxWN = max(lumaN, lumaW);
	float minWN = min(lumaN, lumaW);
	
	float rangeMax = max(maxWN, maxESM);
	float rangeMin = min(minWN, minESM);
	float rangeMaxScaled = rangeMax * fxaaQualityEdgeThreshold;
	float range = rangeMax - rangeMin;
	float rangeMaxClamped = max(fxaaQualityEdgeThresholdMin, rangeMaxScaled);

	bool earlyExit = range < rangeMaxClamped;
	if(earlyExit)
		return rgbyM;

	float lumaNW = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2(-1,-1)));
	float lumaSE = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2( 1, 1)));
	float lumaNE = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2( 1,-1)));
	float lumaSW = FxaaLuma(textureLodOffset(tex, posM, 0.0, ivec2(-1, 1)));
	float lumaNS = lumaN + lumaS;
	float lumaWE = lumaW + lumaE;
	
	float subpixRcpRange = 1.0/range;
	float subpixNSWE = lumaNS + lumaWE;
	float edgeHorz1 = (-2.0 * rgbyM.y) + lumaNS;
	float edgeVert1 = (-2.0 * rgbyM.y) + lumaWE;
	float lumaNESE = lumaNE + lumaSE;
	float lumaNWNE = lumaNW + lumaNE;
	float edgeHorz2 = (-2.0 * lumaE) + lumaNESE;
	float edgeVert2 = (-2.0 * lumaN) + lumaNWNE;
	float lumaNWSW = lumaNW + lumaSW;
	float lumaSWSE = lumaSW + lumaSE;
	float edgeHorz4 = (abs(edgeHorz1) * 2.0) + abs(edgeHorz2);
	float edgeVert4 = (abs(edgeVert1) * 2.0) + abs(edgeVert2);
	float edgeHorz3 = (-2.0 * lumaW) + lumaNWSW;
	float edgeVert3 = (-2.0 * lumaS) + lumaSWSE;
	float edgeHorz = abs(edgeHorz3) + edgeHorz4;
	float edgeVert = abs(edgeVert3) + edgeVert4;
	float subpixNWSWNESE = lumaNWSW + lumaNESE;
	float lengthSign = fxaaQualityRcpFrame.x;
	
	bool horzSpan = edgeHorz >= edgeVert;
	float subpixA = subpixNSWE * 2.0 + subpixNWSWNESE;
	if(!horzSpan) lumaN = lumaW;
	if(!horzSpan) lumaS = lumaE;
	if(horzSpan) lengthSign = fxaaQualityRcpFrame.y;
	float subpixB = (subpixA * (1.0/12.0)) - rgbyM.y;
	float gradientN = lumaN - rgbyM.y;
	float gradientS = lumaS - rgbyM.y;
	float lumaNN = lumaN + rgbyM.y;
	float lumaSS = lumaS + rgbyM.y;
	bool pairN = abs(gradientN) >= abs(gradientS);
	float gradient = max(abs(gradientN), abs(gradientS));
	if(pairN) lengthSign = -lengthSign;
	float subpixC = clamp(abs(subpixB) * subpixRcpRange, 0.0, 1.0);
	
	vec2 posB;
	posB.x = posM.x;
	posB.y = posM.y;		
	vec2 offNP;
	
	offNP.x = (!horzSpan) ? 0.0 : fxaaQualityRcpFrame.x;
	offNP.y = ( horzSpan) ? 0.0 : fxaaQualityRcpFrame.y;	
	if(!horzSpan) posB.x += lengthSign * 0.5;
	if( horzSpan) posB.y += lengthSign * 0.5;
	
	vec2 posN;
	posN.x = posB.x - offNP.x * 1.0;
	posN.y = posB.y - offNP.y * 1.0;
	
	vec2 posP;
	posP.x = posB.x + offNP.x * 1.0;
	posP.y = posB.y + offNP.y * 1.0;
	
	float subpixD = ((-2.0)*subpixC) + 3.0;
	float lumaEndN = FxaaLuma(textureLod(tex, posN, 0.0));
	float subpixE = subpixC * subpixC;
	float lumaEndP = FxaaLuma(textureLod(tex, posP, 0.0));
	
	if(!pairN) lumaNN = lumaSS;
	float gradientScaled = gradient * 1.0/4.0;
	float lumaMM = rgbyM.y - lumaNN * 0.5;
	float subpixF = subpixD * subpixE;
	bool lumaMLTZero = lumaMM < 0.0;
	lumaEndN -= lumaNN * 0.5;
	lumaEndP -= lumaNN * 0.5;
	bool doneN = abs(lumaEndN) >= gradientScaled;
	bool doneP = abs(lumaEndP) >= gradientScaled;
	
	if(!doneN) posN.x -= offNP.x * 1.5;
	if(!doneN) posN.y -= offNP.y * 1.5;
	bool doneNP = (!doneN) || (!doneP);
	if(!doneP) posP.x += offNP.x * 1.5;
	if(!doneP) posP.y += offNP.y * 1.5;
	if(doneNP) {
	if(!doneN) lumaEndN = FxaaLuma(textureLod(tex, posN.xy, 0.0));
	if(!doneP) lumaEndP = FxaaLuma(textureLod(tex, posP.xy, 0.0));
	if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
	if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
	doneN = abs(lumaEndN) >= gradientScaled;
	doneP = abs(lumaEndP) >= gradientScaled;
	
	if(!doneN) posN.x -= offNP.x * 2.0;
	if(!doneN) posN.y -= offNP.y * 2.0;
	doneNP = (!doneN) || (!doneP);
	if(!doneP) posP.x += offNP.x * 2.0;
	if(!doneP) posP.y += offNP.y * 2.0;
	if(doneNP) {
	if(!doneN) lumaEndN = FxaaLuma(textureLod(tex, posN.xy, 0.0));
	if(!doneP) lumaEndP = FxaaLuma(textureLod(tex, posP.xy, 0.0));
	if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
	if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
	doneN = abs(lumaEndN) >= gradientScaled;
	doneP = abs(lumaEndP) >= gradientScaled;
	
	if(!doneN) posN.x -= offNP.x * 2.0;
	if(!doneN) posN.y -= offNP.y * 2.0;
	doneNP = (!doneN) || (!doneP);
	if(!doneP) posP.x += offNP.x * 2.0;
	if(!doneP) posP.y += offNP.y * 2.0;
	if(doneNP) {
	if(!doneN) lumaEndN = FxaaLuma(textureLod(tex, posN.xy, 0.0));
	if(!doneP) lumaEndP = FxaaLuma(textureLod(tex, posP.xy, 0.0));
	if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
	if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
	doneN = abs(lumaEndN) >= gradientScaled;
	doneP = abs(lumaEndP) >= gradientScaled;
	
	if(!doneN) posN.x -= offNP.x * 4.0;
	if(!doneN) posN.y -= offNP.y * 4.0;
	doneNP = (!doneN) || (!doneP);
	if(!doneP) posP.x += offNP.x * 4.0;
	if(!doneP) posP.y += offNP.y * 4.0;
	if(doneNP) {
	if(!doneN) lumaEndN = FxaaLuma(textureLod(tex, posN.xy, 0.0));
	if(!doneP) lumaEndP = FxaaLuma(textureLod(tex, posP.xy, 0.0));
	if(!doneN) lumaEndN = lumaEndN - lumaNN * 0.5;
	if(!doneP) lumaEndP = lumaEndP - lumaNN * 0.5;
	doneN = abs(lumaEndN) >= gradientScaled;
	doneP = abs(lumaEndP) >= gradientScaled;
	
	if(!doneN) posN.x -= offNP.x * 12.0;
	if(!doneN) posN.y -= offNP.y * 12.0;
	doneNP = (!doneN) || (!doneP);
	if(!doneP) posP.x += offNP.x * 12.0;
	if(!doneP) posP.y += offNP.y * 12.0;}}}}	
	float dstN = posM.x - posN.x;
	float dstP = posP.x - posM.x;
	if(!horzSpan) dstN = posM.y - posN.y;
	if(!horzSpan) dstP = posP.y - posM.y;

	bool goodSpanN = (lumaEndN < 0.0) != lumaMLTZero;
	float spanLength = (dstP + dstN);
	bool goodSpanP = (lumaEndP < 0.0) != lumaMLTZero;
	float spanLengthRcp = 1.0/spanLength;
	bool directionN = dstN < dstP;
	float dst = min(dstN, dstP);
	bool goodSpan = directionN ? goodSpanN : goodSpanP;
	float subpixG = subpixF * subpixF;
	float pixelOffset = (dst * (-spanLengthRcp)) + 0.5;
	float subpixH = subpixG * fxaaQualitySubpix;

	float pixelOffsetGood = goodSpan ? pixelOffset : 0.0;
	float pixelOffsetSubpix = max(pixelOffsetGood, subpixH);
	if(!horzSpan) posM.x += pixelOffsetSubpix * lengthSign;
	if( horzSpan) posM.y += pixelOffsetSubpix * lengthSign;

	return vec4(textureLod(tex, posM, 0.0).xyz, rgbyM.y);
}

vec4 FXAAPass(vec4 color, vec2 v_texcoord0)
{
	#if (DebugFxaa == 1)
	color = FxaaPixelShader(v_texcoord0, sampler0, textureSize(sampler0, 0), FxaaSubpixMax, FXAAEdgeThreshold, FXAAEdgeThresholdMin); //Visualize effect (debug)
	#else
	color = FxaaPixelShader(v_texcoord0, sampler0, u_texelDelta.xy, FxaaSubpixMax, FXAAEdgeThreshold, FXAAEdgeThresholdMin);
	#endif
	
	return color;
}
#endif

/*------------------------------------------------------------------------------
                             [NFAA CODE SECTION]
------------------------------------------------------------------------------*/

#if (NFAA_PASS == 1)
float filterStrength = FilterStrength;
float filterSpread = FilterSpread;

float GetColorLuminance(vec3 color)
{
	vec3 lumCoeff = vec3(0.212656f, 0.715158f, 0.072186f);
	return dot(color.rgb, lumCoeff);
}

vec4 textureBiLinear(sampler2D tex, vec2 v_texcoord0)
{	
	float texelSizeX = (1.0 / u_texelDelta.x); //size of one texel 
	float texelSizeY = (1.0 / u_texelDelta.y); //size of one texel 

	int nX = int(v_texcoord0.x * u_texelDelta.x);
	int nY = int(v_texcoord0.y * u_texelDelta.y);
	
	vec2 texCoord_New = vec2((float(nX) + 0.5) / u_texelDelta.x,
							  (float(nY) + 0.5) / u_texelDelta.y);
	// Take nearest two data in current row.
    vec4 p0q0 = texture(tex, texCoord_New);
    vec4 p1q0 = texture(tex, texCoord_New + vec2(texelSizeX, 0));

	// Take nearest two data in bottom row.
    vec4 p0q1 = texture(tex, texCoord_New + vec2(0, texelSizeY));
    vec4 p1q1 = texture(tex, texCoord_New + vec2(texelSizeX , texelSizeY));

    float a = fract(v_texcoord0.x * u_texelDelta.x); // Get Interpolation factor for X direction.

	// Interpolation in X direction.
    vec4 pInterp_q0 = mix(p0q0, p1q0, a); // Interpolates top row in X direction.
    vec4 pInterp_q1 = mix(p0q1, p1q1, a); // Interpolates bottom row in X direction.

    float b = fract(v_texcoord0.y * u_texelDelta.y); // Get Interpolation factor for Y direction.
	
    return mix(pInterp_q0, pInterp_q1, b); // Interpolate in Y direction.
}

vec2 findContrastByLuminance(vec2 v_texcoord0) 
{
	// Normal offsets, scale by filter spread
	vec2 upOffset = vec2(0, u_texelDelta.y) * filterSpread;
	vec2 rightOffset = vec2(u_texelDelta.x, 0) * filterSpread;

	float topHeight = GetColorLuminance(texture( sampler0, v_texcoord0 + upOffset).rgb);
	float bottomHeight = GetColorLuminance(texture( sampler0, v_texcoord0 - upOffset).rgb);
	float rightHeight = GetColorLuminance(texture( sampler0, v_texcoord0 + rightOffset).rgb);
	float leftHeight = GetColorLuminance(texture( sampler0, v_texcoord0 - rightOffset).rgb);
	float leftTopHeight = GetColorLuminance(texture( sampler0, v_texcoord0 - rightOffset + upOffset).rgb);
	float leftBottomHeight = GetColorLuminance(texture( sampler0, v_texcoord0 - rightOffset - upOffset).rgb);
	float rightBottomHeight = GetColorLuminance(texture( sampler0, v_texcoord0 + rightOffset + upOffset).rgb);
	float rightTopHeight = GetColorLuminance(texture( sampler0, v_texcoord0 + rightOffset - upOffset).rgb);

	// Normal map creation
	float sum0 = rightTopHeight    + bottomHeight + leftTopHeight;
	float sum1 = leftBottomHeight  + topHeight    + rightBottomHeight;
	float sum2 = leftTopHeight     + rightHeight  + leftBottomHeight;
	float sum3 = rightBottomHeight + leftHeight   + rightTopHeight;

	// Subtract the opposite sample set for final vectors
	float vector1 = (sum0 - sum1) * filterStrength;
	float vector2 = (sum3 - sum2) * filterStrength;

	vec2 Vectors = vec2(vector1, vector2);

	return Vectors;
}

vec2 findContrastByColor(vec2 v_texcoord0)
{
	// Normal offsets, scale by filter spread
	vec2 upOffset = vec2(0, u_texelDelta.x) * filterSpread;
	vec2 rightOffset = vec2(u_texelDelta.x, 0) * filterSpread;

	vec3 topHeight = texture( sampler0, v_texcoord0 + upOffset).rgb;
	vec3 bottomHeight = texture( sampler0, v_texcoord0 - upOffset).rgb;
	vec3 rightHeight = texture( sampler0, v_texcoord0 + rightOffset).rgb;
	vec3 leftHeight = texture( sampler0, v_texcoord0 - rightOffset).rgb;
	vec3 leftTopHeight = texture( sampler0, v_texcoord0 - rightOffset + upOffset).rgb;
	vec3 leftBottomHeight = texture( sampler0, v_texcoord0 - rightOffset - upOffset).rgb;
	vec3 rightBottomHeight = texture( sampler0, v_texcoord0 + rightOffset + upOffset).rgb;
	vec3 rightTopHeight = texture( sampler0, v_texcoord0 + rightOffset - upOffset).rgb;

	// Normal map creation
	vec3 sum0 = rightTopHeight    + bottomHeight + leftTopHeight;
	vec3 sum1 = leftBottomHeight  + topHeight    + rightBottomHeight;
	vec3 sum2 = leftTopHeight     + rightHeight  + leftBottomHeight;
	vec3 sum3 = rightBottomHeight + leftHeight   + rightTopHeight;

	// Subtract the opposite sample set for final vectors
	float vector1 = length(sum0 - sum1) * filterStrength;
	float vector2 = length(sum3 - sum2) * filterStrength;

	vec2 Vectors = vec2(vector1, vector2);

	return Vectors;
}

vec4 NFAAPass(vec4 color, vec2 v_texcoord0) 
{
	vec2 Vectors;
	filterStrength += (filterSpread/2);
	
	#if (UseColor == 1) 
		Vectors = findContrastByColor(v_texcoord0.xy);
	#else
		Vectors = findContrastByLuminance(v_texcoord0.xy);
	#endif

	float filterClamp = filterStrength/filterSpread;

	Vectors.xy = clamp(Vectors, -vec2(filterClamp,filterClamp), vec2(filterClamp,filterClamp));

	vec2 Normal = vec2(Vectors.x, Vectors.y) * u_texelDelta.xy;	
	
	#if (BilinearSampling == 1)
	vec4 Scene0 = textureBiLinear(sampler0, v_texcoord0.xy);
	vec4 Scene1 = textureBiLinear(sampler0, v_texcoord0.xy + Normal.xy);
	vec4 Scene2 = textureBiLinear(sampler0, v_texcoord0.xy - Normal.xy);
	vec4 Scene3 = textureBiLinear(sampler0, v_texcoord0.xy + vec2(Normal.x, -Normal.y));
	vec4 Scene4 = textureBiLinear(sampler0, v_texcoord0.xy - vec2(Normal.x, -Normal.y));
	#else
	vec4 Scene0 = texture(sampler0, v_texcoord0.xy, 0.0);
	vec4 Scene1 = texture(sampler0, v_texcoord0.xy + Normal.xy, 0.0);
	vec4 Scene2 = texture(sampler0, v_texcoord0.xy - Normal.xy, 0.0);
	vec4 Scene3 = texture(sampler0, v_texcoord0.xy + vec2(Normal.x, -Normal.y), 0.0);
	vec4 Scene4 = texture(sampler0, v_texcoord0.xy - vec2(Normal.x, -Normal.y), 0.0);
	#endif

	// Final color
	color = (Scene0 + Scene1 + Scene2 + Scene3 + Scene4) * 0.2f;

	// Debug Output
	#if (DebugNFAA == 1) 
		color.xyz = normalize(vec3(Vectors.x, Vectors.y, 1.0) * 0.5 + 0.5);
	#endif

	return color;
}
#endif

/*------------------------------------------------------------------------------
                        [TEXTURE FILTERING CODE SECTION]
------------------------------------------------------------------------------*/

#if (BICUBIC_FILTERING == 1)
float BSpline(float x)
{
	float f = x;
	
	if(f < 0.0)
	{
		f = -f;
	}
	if(f >= 0.0 && f <= 1.0)
	{
		return (2.0 / 3.0) + (0.5) * (f* f * f) - (f*f);
	}
	else if(f > 1.0 && f <= 2.0)
	{
		return 1.0 / 6.0 * pow(( 2.0 - f  ), 3.0);
	}
	return 1.0;
}

float CatMullRom(float x)
{
    float B = 0.0;
    float C = 0.5;
    float f = x;
    if( f < 0.0 )
    {
        f = -f;
    }
    if( f < 1.0 )
    {
        return ((12.0 - 9.0 * B - 6.0 * C) * (f * f * f) +
            (-18.0f + 12.0f * B + 6.0f * C) * (f * f) +
            (6.0f - 2.0f * B) ) / 6.0f;
    }
    else if( f >= 1.0 && f < 2.0 )
    {
        return (( -B - 6 * C) * (f * f * f)
            + (6 * B + 30 * C) * (f *f) +
            (- (12 * B) - 48 * C ) * f +
            8 * B + 24 * C)/ 6.0;
    }
    else
    {
        return 0.0;
    }
}

float Bell(float x)
{
	float f = (x / 2.0) * 1.5;
	if(f > -1.5 && f < -0.5)
	{
		return(0.5 * pow(f + 1.5, 2.0));
	}
	else if(f > -0.5 && f < 0.5)
	{
		return 3.0 / 4.0 - (f * f);
	}
	else if(( f > 0.5 && f < 1.5 ))
	{
		return(0.5 * pow(f - 1.5, 2.0));
	}
	return 0.0;
}

float Triangular(float f)
{
	f = f / 2.0;
	if(f < 0.0)
	{
		return (f + 1.0);
	}
	else
	{
		return (1.0 - f);
	}
	return 0.0;
}

vec4 BiCubicFilter(sampler2D tex, vec2 v_texcoord0)
{
	float texelSizeX = (1.0/u_texelDelta.x); //size of one texel 
	float texelSizeY = (1.0/u_texelDelta.y); //size of one texel
	
    vec4 nSum = vec4(0.0, 0.0, 0.0, 0.0);
    vec4 nDenom = vec4(0.0, 0.0, 0.0, 0.0);
	
    float a = fract(v_texcoord0.x * u_texelDelta.x); // get the decimal part
    float b = fract(v_texcoord0.y * u_texelDelta.y); // get the decimal part

	int nX = int(v_texcoord0.x * u_texelDelta.x);
	int nY = int(v_texcoord0.y * u_texelDelta.y);
	
	vec2 TexCoord1 = vec2( float(nX) / u_texelDelta.x + OffsetAmount / u_texelDelta.x,
					       float(nY) / u_texelDelta.y + OffsetAmount / u_texelDelta.y );

    for(int m = -1; m <= 2; m++)
    {
        for( int n = -1; n <= 2; n++)
        {
			vec4 Sample = texture(tex, TexCoord1 + vec2(texelSizeX * float(m), texelSizeY * float(n)));
			float f  = FilterType(float(m) - a);
			vec4 vecCooef1 = vec4( f,f,f,f );
			float f1 = FilterType(-(float(n) - b));
			vec4 vecCoeef2 = vec4( f1, f1, f1, f1 );
            nSum = nSum + (Sample * vecCoeef2 * vecCooef1);
            nDenom = nDenom + ((vecCoeef2 * vecCooef1));
        }
    }
    return nSum / nDenom;
}

vec4 BiCubicPass(vec4 color, vec2 v_texcoord0)
{
   color = BiCubicFilter(sampler0, v_texcoord0);
   return color;
}
#endif

/*------------------------------------------------------------------------------
						[GAMMA CORRECTION CODE SECTION]
------------------------------------------------------------------------------*/

#if (GAMMA_CORRECTION == 1)
vec4 PostGammaPass(vec4 color, vec2 v_texcoord0)
{
	float gamma = Gamma;
	vec3 correctedGamma;
	
	correctedGamma.r = InverseGammaCorrection(color.r, (gamma/2.2f));
	correctedGamma.g = InverseGammaCorrection(color.g, (gamma/2.2f));
	correctedGamma.b = InverseGammaCorrection(color.b, (gamma/2.2f));
	
	color.rgb = correctedGamma.rgb;
	color.a = RGBLuminance(color.rgb);

	return color;
}
#endif

/*------------------------------------------------------------------------------
                          [LUMA SHARPEN CODE SECTION]
------------------------------------------------------------------------------*/

#if (TEXTURE_SHARPENING == 1)
#define px (1.0/textureSize(sampler0, 0)).x
#define py (1.0/textureSize(sampler0, 0)).y
#define CoefLuma vec3(0.212656f, 0.715158f, 0.072186f)
#define sharpenLumaStrength (CoefLuma * SharpenStrength)

vec4 TexSharpenPass(vec4 color, vec2 v_texcoord0) 
{
	vec3 ColOffset;
	vec3 Pix = texture(sampler0, v_texcoord0).rgb; 

	ColOffset = texture(sampler0, v_texcoord0 + vec2(-px,py) * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(px,-py) * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(-px,-py)  * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(px,py) * OffsetBias).rgb;

	ColOffset += Pix.rgb;

	ColOffset += texture(sampler0, v_texcoord0 + vec2(0,py) * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(0,-py) * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(-px,0) * OffsetBias).rgb;
	ColOffset += texture(sampler0, v_texcoord0 + vec2(px,0) * OffsetBias).rgb;

	ColOffset /= 9.0f;

	vec3 sharp = Pix - ColOffset;
	
	float sharpenLuma = dot(sharp, sharpenLumaStrength);
	sharpenLuma = clamp(sharpenLuma, -SharpenClamp, SharpenClamp);

	color.rgb = color.rgb + sharpenLuma;
	color.a = RGBLuminance(color.rgb);
	
	#if (DebugSharpen == 1)
		color = vec4((0.5f + (sharpenLuma * 4))).rrrr; //visualise sharpening (for debugging)
	#endif
	
	return clamp(color, 0.0, 1.0);
}
#endif

/*------------------------------------------------------------------------------
							[VIBRANCE CODE SECTION]
------------------------------------------------------------------------------*/

#if (PIXEL_VIBRANCE == 1)
vec4 VibrancePass(vec4 color, vec2 v_texcoord0)
{
	float luma = RGBLuminance(color.rgb);
	
	float colorMax = max(color.r, max(color.g, color.b)); //Find the strongest color
	float colorMin = min(color.r, min(color.g, color.b)); //Find the weakest color
	float colorSaturation = colorMax - colorMin; //The difference between the two is the saturation

	color.rgb = mix(vec3(luma), color.rgb, (1.0 + (Vibrance * (1.0 - (sign(Vibrance) * colorSaturation))))); //extrapolate
	color.a = RGBLuminance(color.rgb);

	return clamp(color, 0.0, 1.0); //Debug: return colorSaturation.xxxx;
}
#endif

/*------------------------------------------------------------------------------
						   [BLOOM PASS CODE SECTION]
------------------------------------------------------------------------------*/

#if (BLOOM_PASS == 1)
const float blurspread = BlurSpread;
const float bloomPower = BloomPower;

vec3 BlendScreen(vec3 base, vec3 blend)
{
	if (BlendPower == 1)
		return (base + blend) - (base * blend);
	else
		return (blend + blend) - (blend * blend);
}

vec3 BlendAddLight(vec3 base, vec3 blend)
{
	return base + blend;
}

vec3 BlendOverlay(vec3 base, vec3 blend)
{
	return vec3((blend.x <= 0.5) ? (2.0 * base.x * blend.x) 
			: (1.0 - 2.0 * (1.0 - blend.x) * (1.0 - base.x)),
			(blend.y <= 0.5) ? (2.0 * base.y * blend.y) 
			: (1.0 - 2.0 * (1.0 - blend.y) * (1.0 - base.y)),
			(blend.z <= 0.5) ? (2.0 * base.z * blend.z) 
			: (1.0 - 2.0 * (1.0 - blend.z) * (1.0 - base.z)));
}

vec4 BloomPass(vec4 color, vec2 v_texcoord0)
{	
	vec4 blur = texture(sampler0, v_texcoord0);
	blur = color * 0.22520613262190495;

	vec2 dx = vec2(1.0/textureSize(sampler0, 0).x * blurspread, u_texelDelta.x);
	vec2 dy = vec2(u_texelDelta.y, 1.0/textureSize(sampler0, 0).y * blurspread);
	vec2 dx2 = 2 * dx;
	vec2 dy2 = 2 * dy;

	vec4 colorU = texture(sampler0, v_texcoord0 + dy);

	blur += 0.002589001911021066 * texture(sampler0, v_texcoord0 -dx2 +dy2); 
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 -dx +dy2); 
	blur += 0.024146616900339800 * texture(sampler0, v_texcoord0 +dy2); 
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 +dx +dy2); 
	blur += 0.002589001911021066 * texture(sampler0, v_texcoord0 +dx2 +dy2); 

	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 -dx2 +dy);
	blur += 0.044875475183061630 * texture(sampler0, v_texcoord0 -dx +dy);
	blur += 0.100529757860782610 * texture(sampler0, v_texcoord0 +dy);
	blur += 0.044875475183061630 * texture(sampler0, v_texcoord0 +dx +dy);
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 +dx2 +dy);

	blur += 0.024146616900339800 * texture(sampler0, v_texcoord0 -dx2);
	blur += 0.100529757860782610 * texture(sampler0, v_texcoord0 -dx);	
	blur += 0.100529757860782610 * texture(sampler0, v_texcoord0 +dx);
	blur += 0.024146616900339800 * texture(sampler0, v_texcoord0 +dx2);

	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 -dx2 -dy);
	blur += 0.044875475183061630 * texture(sampler0, v_texcoord0 -dx -dy);
	blur += 0.100529757860782610 * texture(sampler0, v_texcoord0 -dy);
	blur += 0.044875475183061630 * texture(sampler0, v_texcoord0 +dx -dy);
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 +dx2 -dy);

	blur += 0.002589001911021066 * texture(sampler0, v_texcoord0 -dx2 -dy2);
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 -dx -dy2);
	blur += 0.024146616900339800 * texture(sampler0, v_texcoord0 -dy2);
	blur += 0.010778807494659370 * texture(sampler0, v_texcoord0 +dx -dy2);
	blur += 0.002589001911021066 * texture(sampler0, v_texcoord0 +dx2 -dy2);				

	vec4 newcolor;
	newcolor.rgb = BloomType(color.rgb, blur.rgb);
	newcolor.a = RGBLuminance(newcolor.rgb);
	color.a = RGBLuminance(color.rgb);
	
	#if (BloomMixType == 1)
	color = mix(color, newcolor, bloomPower);
	#endif
	#if (BloomMixType == 2)
	color = (mix(color, newcolor, bloomPower) + mix(newcolor, blur, bloomPower))/2;
	#endif
	#if (BloomMixType == 3)
	color = mix(color, newcolor, mix(color.r*0.5, newcolor.r, bloomPower));
	#endif
	
	return color;
}
#endif

/*------------------------------------------------------------------------------
               [COLOR CORRECTION/TONE MAPPING PASS CODE SECTION]
------------------------------------------------------------------------------*/

#if (SCENE_TONEMAPPING == 1)
float GetLuminance (vec3 rgb)
{
	return (0.212656f * rgb.x) + (0.715158f * rgb.y) + (0.072186f * rgb.z);
}

vec4 TonemapPass(vec4 color, vec2 v_texcoord0)
{

	float luminance = GetLuminance(color.rgb);
	
	// Average scene luminance is stored in R channel
	#if (TonemapType == 2)
		float avgLuminance = clamp(color.r, ToneAmount, 1.0);
	#else
		float avgLuminance = Luma;
	#endif
	
	float exposure = Exposure;
	float whitePoint = WhitePoint;
	float blackLevel = ToneAmount;

	//Scene tone mapping
	#if (TonemapType == 1)
		color.r = InverseGammaCorrection(color.r, (blackLevel));
		color.g = InverseGammaCorrection(color.g, (blackLevel));
		color.b = InverseGammaCorrection(color.b, (blackLevel));
	#endif

	vec3 lumCoeff = vec3(0.212656f, 0.715158f, 0.072186f);
	
	// RGB -> XYZ conversion
	const mat3 RGB2XYZ = mat3(0.4124564f,  0.3575761f,  0.1804375f,
							  0.2126729f,  0.7151522f,  0.0721750f,
							  0.0193339f,  0.1191920f,  0.9503041f);
							  
	vec3 XYZ = (color.rgb*RGB2XYZ);
  
	// XYZ -> Yxy conversion
	vec3 Yxy = lumCoeff;

	Yxy.r = XYZ.g;                            // copy luminance Y
	Yxy.g = XYZ.r / (XYZ.r + XYZ.g + XYZ.b ); // x = X / (X + Y + Z)
	Yxy.b = XYZ.g / (XYZ.r + XYZ.g + XYZ.b ); // y = Y / (X + Y + Z)
	
	// (Lp) Map average luminance to the middlegrey zone by scaling pixel luminance
	#if (TonemapType == 1)
		float Lp = Yxy.r * exposure / avgLuminance;
	#endif
	#if (TonemapType == 2)
		float Lp = (Yxy.r * avgLuminance) * exposure / 100.0f;  
	#endif
	#if (TonemapType == 3)
		float Lp = ((Yxy.r * luminance) + (Yxy.g * luminance) + (Yxy.b * luminance)) * ((exposure / avgLuminance)*blackLevel);  
	#endif
	
	// (Ld) Scale all luminance within a displayable range of 0 to 1
	Yxy.r = (Lp * (1.0f + Lp/(whitePoint * whitePoint)))/(1.0f + Lp);
  
	// Yxy -> XYZ conversion
	XYZ.r = Yxy.r * Yxy.g / Yxy. b;               // X = Y * x / y
	XYZ.g = Yxy.r;                                // copy luminance Y
	XYZ.b = Yxy.r * (1.0f - Yxy.g - Yxy.b) / Yxy.b;  // Z = Y * (1-x-y) / y
	
	// XYZ -> RGB conversion
	const mat3 XYZ2RGB  = mat3(3.2404542f, -1.5371385f, -0.4985314f,
							  -0.9692660f,  1.8760108f,  0.0415560f, 
							   0.0556434f, -0.2040259f,  1.0572252f);

	color.rgb = clamp(XYZ * XYZ2RGB, 0.0, 1.0);
	color.a = RGBLuminance(color.rgb);
	
	return color;
}
#endif

/*------------------------------------------------------------------------------
						[S_CURVE CONTRAST CODE SECTION]
------------------------------------------------------------------------------*/

#if (S_CURVE_CONTRAST == 1)
vec4 CurvesPass(vec4 color, vec2 v_texcoord0)
{
	float CurveBlend = CurvesContrast;

	#if (CurveType != 2)
		float luma = RGBLuminance(color.rgb);
		vec3 chroma = color.rgb - luma;
	#endif

	#if (CurveType == 2)
		vec3 x = color.rgb;
	#elif (CurveType == 1)
		vec3 x = chroma;
		x = x * 0.5 + 0.5;
	#else
		float x = luma;
	#endif

	// -- Curve 1 -- Cubic Bezier spline
	#if (CurvesFormula == 1)
		float a = 0.00; //start point
		float b = 0.00; //control point 1
		float c = 1.00; //control point 2
		float d = 1.00; //endpoint

		float r  = (1-x);
		float r2 = r*r;
		float r3 = r2 * r;
		float x2 = x*x;
		float x3 = x2*x;

		x = a*(1-x)*(1-x)*(1-x) + 3*b*(1-x)*(1-x)*x + 3*c*(1-x)*x*x + d*x*x*x;
	#endif

	// -- Curve 2 -- Cubic Bezier spline II
	#if (CurvesFormula == 2)
		vec3 a = vec3(0.00f,0.00f,0.00f); //start point
		vec3 b = vec3(0.25f,0.15f,0.85f); //control point 1
		vec3 c = vec3(0.75f,0.85f,0.15f); //control point 2
		vec3 d = vec3(1.00f,1.00f,1.00f); //endpoint

		vec3 ab = mix(a,b,x);           // point between a and b (green)
		vec3 bc = mix(b,c,x);           // point between b and c (green)
		vec3 cd = mix(c,d,x);           // point between c and d (green)
		vec3 abbc = mix(ab,bc,x);       // point between ab and bc (blue)
		vec3 bccd = mix(bc,cd,x);       // point between bc and cd (blue)
		vec3 dest = mix(abbc,bccd,x);   // point on the bezier-curve (black)
		x = dest;
	#endif

	#if (CurveType == 2) //Both Luma and Chroma
		vec3 LColor = x; 
		color.rgb = mix(color.rgb, LColor, CurveBlend);
	#elif (CurveType == 1) //Only Chroma
		x = x * 2 - 1;
		vec3 LColor = luma + x; //Luma + Chroma
		color.rgb = mix(color.rgb, LColor, CurveBlend);
	#else // CurveType == 0 //Only Luma
		x = mix(luma, x, CurveBlend);
		color.rgb = x + chroma; //Luma + Chroma 
		color.a = RGBLuminance(color.rgb);
	#endif
	
	return color;
}
#endif

/*------------------------------------------------------------------------------
							[TOON SHADING CODE SECTION]
------------------------------------------------------------------------------*/

#if (TOON_SHADING == 1)
vec4 CelPass(vec4 color, vec2 v_texcoord0)
{
	#define pixel vec2(1.0/textureSize(sampler0, 0).x, 1.0/textureSize(sampler0, 0).y)
	
	vec3 ToonLumaCoef = vec3(0.2126, 0.7152, 0.0722);
	float Tluma = dot(color.rgb, ToonLumaCoef);
	
	float Tmax_color = max(color.r, max(color.g,color.b));
	float Tmin_color = min(color.r, min(color.g,color.b));

	float Tcolor_saturation = Tmax_color - Tmin_color;

	color.rgb = mix(vec3(Tluma), color.rgb, (1.0 + (ToonVibrance * (1.0-(sign(ToonVibrance) * Tcolor_saturation)))));

	float diff1 = dot(ToonLumaCoef, texture(sampler0, v_texcoord0 + pixel).rgb);
	diff1 = dot(vec4(ToonLumaCoef, -1.0), vec4(texture(sampler0, v_texcoord0 - pixel).rgb, diff1));
	
	float diff2 = dot(ToonLumaCoef, texture(sampler0, v_texcoord0 +vec2(pixel.x,-pixel.y)).rgb);
	diff2 = dot(vec4(ToonLumaCoef, -1.0), vec4(texture(sampler0, v_texcoord0 +vec2(-pixel.x,pixel.y)).rgb, diff2));

	float edge = dot(vec2(diff1, diff2), vec2(diff1, diff2));

	color.rgb = pow(edge, ToonEdgeSlope) * -ToonStrength + color.rgb;
	color.a = RGBLuminance(color.rgb);

	return color;
}
#endif

/*------------------------------------------------------------------------------
                      [MAIN() & COMBINE PASS CODE SECTION]
------------------------------------------------------------------------------*/

void main()
{
	#if (NFAA_PASS == 0 && FXAA_PASS == 0 && BICUBIC_FILTERING == 0)
		gl_FragColor = texture(sampler0, v_texcoord0);
		gl_FragColor.a = RGBLuminance(gl_FragColor.rgb);
	#endif
	
	#if (FXAA_PASS == 1)
		gl_FragColor = PreFxaaPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (BICUBIC_FILTERING)
		gl_FragColor = BiCubicPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (NFAA_PASS == 1)
		gl_FragColor = NFAAPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (FXAA_PASS == 1)
		gl_FragColor = FXAAPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (TEXTURE_SHARPENING == 1)
		gl_FragColor = TexSharpenPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (GAMMA_CORRECTION == 1)
		gl_FragColor = PostGammaPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (PIXEL_VIBRANCE == 1)
		gl_FragColor = VibrancePass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (BLOOM_PASS == 1)
		gl_FragColor = BloomPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (SCENE_TONEMAPPING == 1)
		gl_FragColor = TonemapPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (S_CURVE_CONTRAST == 1)
		gl_FragColor = CurvesPass(gl_FragColor, v_texcoord0);
	#endif
	
	#if (TOON_SHADING == 1)
		gl_FragColor = CelPass(gl_FragColor, v_texcoord0);
	#endif
}
